import logging
from ...utils import get_gpu_count, get_cpu_count

__all__ = ['Resources', 'DistributedResource',
           'get_remote_cpu_count', 'get_remote_gpu_count',
           'get_gpu_count', 'get_cpu_count']

logger = logging.getLogger(__name__)


class Resources(object):
    """Resource for AutoGluon Scheduler :class:`autogluon.scheduler.TaskScheduler`

    Args:
        num_cpus (int): number of cpu cores required for the training task.
        num_gpus (int): number of gpu required for the training task.

    Example:
        >>> def my_task():
        >>>     pass
        >>> resource = Resources(num_cpus=2, num_gpus=0)
        >>> task = Task(my_task, {}, resource)
    """
    def __init__(self, num_cpus=1, num_gpus=0):
        self.num_cpus = num_cpus
        self.num_gpus = num_gpus
        self.cpu_ids = []
        self.gpu_ids = []
        self.ready = False

    def _release(self):
        self.ready = False
        self.cpu_ids = []
        self.gpu_ids = []

    def _ready(self, cids, gids):
        self.cpu_ids = cids
        self.gpu_ids = gids
        self.ready = True

    def __repr__(self):
        reprstr = self.__class__.__name__ + '(' \
            + 'nCPUs = ' + str(self.num_cpus)
        if len(self.cpu_ids) > 0:
            reprstr += ', CPU_IDs = {' + str(self.cpu_ids) + '}'
        if self.num_gpus > 0:
            reprstr += ', nGPUs = ' + str(self.num_gpus)
        if len(self.gpu_ids) > 0:
            reprstr += ', GPU_IDs = {' + str(self.gpu_ids) + '}'
        reprstr += ')'
        return reprstr

class DistributedResource(Resources):
    """Resource for AutoGluon Distributed Scheduler :class:`autogluon.distributed.DistributedTaskScheduler`

    Args:
        num_cpus (int): number of cpu cores required for the training task.
        num_gpus (int): number of gpu required for the training task.

    Example:
        >>> def my_task():
        >>>     pass
        >>> resource = DistributedResource(num_cpus=2, num_gpus=1)
        >>> task = Task(my_task, {}, resource)
    """
    def __init__(self, num_cpus=1, num_gpus=0):
        super(DistributedResource, self).__init__(num_cpus, num_gpus)
        self.node = None
        self.is_ready = False

    def _ready(self, remote, cids, gids):
        super(DistributedResource, self)._ready(cids, gids)
        self.node = remote
        self.is_ready = True
 
    def _release(self):
        super(DistributedResource, self)._release()
        self.node = None

    def __repr__(self):
        reprstr = self.__class__.__name__ + '(\n\t'
        if self.node: reprstr  += 'Node = ' + str(self.node)
        reprstr  += '\n\tnCPUs = ' + str(self.num_cpus)
        if len(self.cpu_ids) > 0:
            reprstr += ', CPU_IDs = {' + str(self.cpu_ids) + '}'
        if self.num_gpus > 0:
            reprstr += ', nGPUs = ' + str(self.num_gpus)
        if len(self.gpu_ids) > 0:
            reprstr += ', GPU_IDs = {' + str(self.gpu_ids) + '}'
        reprstr += ')'
        return reprstr


def get_remote_cpu_count(node):
    ret = node.submit(get_cpu_count)
    return ret.result()


def get_remote_gpu_count(node):
    ret = node.submit(get_gpu_count)
    return ret.result()



import logging
import numpy as np
import copy
from typing import NamedTuple, Dict

logger = logging.getLogger(__name__)


class RungEntry(NamedTuple):
    level: int  # Rung level r_j
    prom_quant: float  # Promotion quantile q_j
    data: Dict  # Data of all previous jobs reaching the level


class StoppingRungSystem(object):
    """
    Implements stopping rule resembling the median rule. Once a config is
    stopped, it cannot be promoted later on.
    This is different to what has been published as ASHA (see
    :class:`PromotionRungSystem`).
    """
    def __init__(self, rung_levels, promote_quantiles, max_t):
        # The data entry in _rungs is a dict mapping task_key to
        # reward_value
        assert len(rung_levels) == len(promote_quantiles)
        self._rungs = [
            RungEntry(level=x, prom_quant=y, data=dict())
            for x, y in reversed(list(zip(rung_levels, promote_quantiles)))]

    def on_task_schedule(self):
        return dict()

    def on_task_add(self, task, skip_rungs, **kwargs):
        pass

    @staticmethod
    def _cutoff(recorded, prom_quant):
        if not recorded:
            return None
        return np.percentile(list(recorded.values()), (1 - prom_quant) * 100)

    def on_task_report(self, task, cur_iter, cur_rew, skip_rungs):
        """
        Decision on whether task may continue (task_continues = True), or should be
        stopped (task_continues = False).
        milestone_reached is a flag whether cur_iter coincides with a milestone.
        If True, next_milestone is the next milestone after cur_iter, or None
        if there is none.

        :param task: Only need task.task_id
        :param cur_iter: Current time_attr value of task
        :param cur_rew: Current reward_attr value of task
        :param skip_rungs: This number of lowest rung levels are not
            considered milestones for this task
        :return: dict(task_continues, milestone_reached, next_milestone)
        """
        assert cur_rew is not None, \
            "Reward attribute must be a numerical value, not None"
        task_key = str(task.task_id)
        task_continues = True
        milestone_reached = False
        next_milestone = None
        if skip_rungs > 0:
            milestone_rungs = self._rungs[:(-skip_rungs)]
        else:
            milestone_rungs = self._rungs
        for rung in milestone_rungs:
            milestone = rung.level
            prom_quant = rung.prom_quant
            recorded = rung.data
            if not (cur_iter < milestone or task_key in recorded):
                # Note: It is important for model-based searchers that
                # milestones are reached exactly, not jumped over. In
                # particular, if a future milestone is reported via
                # register_pending, its reward value has to be passed
                # later on via update.
                assert cur_iter == milestone, \
                    "cur_iter = {} > {} = milestone. Make sure to report time attributes covering all milestones".format(
                        cur_iter, milestone)
                milestone_reached = True
                cutoff = self._cutoff(recorded, prom_quant)
                if cutoff is not None and cur_rew < cutoff:
                    task_continues = False
                recorded[task_key] = cur_rew
                break
            next_milestone = milestone
        return {
            'task_continues': task_continues,
            'milestone_reached': milestone_reached,
            'next_milestone': next_milestone}

    def on_task_remove(self, task):
        pass

    def get_first_milestone(self, skip_rungs):
        return self._rungs[-(skip_rungs + 1)].level

    def get_milestones(self, skip_rungs):
        if skip_rungs > 0:
            milestone_rungs = self._rungs[:(-skip_rungs)]
        else:
            milestone_rungs = self._rungs
        return [x.level for x in milestone_rungs]

    def snapshot_rungs(self, skip_rungs):
        if skip_rungs > 0:
            _rungs = self._rungs[:(-skip_rungs)]
        else:
            _rungs = self._rungs
        return [(x.level, x.data) for x in _rungs]

    def __repr__(self):
        iters = " | ".join([
            "Iter {:.3f}: {}".format(
                r.level, self._cutoff(r.data, r.prom_quant))
            for r in self._rungs])
        return "Rung system: " + iters



import numpy as np
from abc import ABC, abstractmethod
from typing import Dict, Tuple, Optional, NamedTuple

from ..tuning_algorithms.base_classes import SurrogateModel, AcquisitionFunction, \
    OutputSurrogateModel, assign_active_metric, dictionarize_objective


class HeadWithGradient(NamedTuple):
    hvals: np.array
    dh_dmean: Dict[str, np.array]  # Maps each output model to its head gradient wrt to the predictive mean
    dh_dstd: Dict[str, np.array]  # Maps each output model to its head gradient wrt to the predictive std


class MeanStdAcquisitionFunction(AcquisitionFunction, ABC):
    """
    Base class for standard acquisition functions which depend on predictive
    mean and stddev. Subclasses have to implement the head and its derivatives
    w.r.t. mean and std:

        f(x, model) = h(mean, std, model.current_best())

    If model is a SurrogateModel, then active_metric is ignored. If model is a Dict mapping output names to models,
    then active_metric must be given.

    NOTE that acquisition functions will always be *minimized*!

    """
    def __init__(self, model: OutputSurrogateModel, active_metric: str = None):
        super().__init__(model, active_metric)
        if isinstance(model, SurrogateModel):
            # Ignore active_metric
            model = dictionarize_objective(model)
        assert isinstance(model, Dict)
        self.model = model
        self.model_output_names = sorted(self.model.keys())
        for output_model in self.model.values():
            assert 'mean' in output_model.keys_predict()
            assert 'std' in output_model.keys_predict()
        self.active_metric = assign_active_metric(self.model, active_metric)

    def compute_acq(self, inputs: np.ndarray, model: Optional[OutputSurrogateModel] = None) -> np.ndarray:
        if model is None:
            model = self.model
        if isinstance(model, SurrogateModel):
            self.model = dictionarize_objective(model)
        if inputs.ndim == 1:
            inputs = inputs.reshape((1, -1))
        output_to_predictions = self._map_outputs_to_predictions(inputs)
        self.num_mcmc_samples = len(output_to_predictions[self.active_metric])  # a single sample means no MCMC
        current_best_list = self._get_current_best_for_active_metric()

        fvals_list = []  # this will contain the acquisition function for each MCMC sample (if any)
        # Loop over MCMC samples
        for predictions, current_best in zip(
                zip(*output_to_predictions.values()),
                current_best_list):
            # Dictionaries mapping each output model name to predictions and current_bests for the current MCMC sample
            output_model_to_mean_std = dict(zip(self.model_output_names, predictions))
            # Create a dictionary mapping each output model name to means, stds and current best,
            # all three in a shape that can be fed into self._compute_heads
            output_to_mean_std = {}
            for output_model, prediction in output_model_to_mean_std.items():
                means = prediction['mean']
                stds = prediction['std']
                num_fantasies = means.shape[1] if means.ndim == 2 else 1
                if num_fantasies > 1:
                    stds = stds.reshape((-1, 1))
                    if current_best is not None:
                        assert current_best.size == num_fantasies, \
                            "mean.shape[1] = {}, current_best.size = {} (must be the same)".format(
                                num_fantasies, current_best.size)
                        current_best = current_best.reshape((1, -1))
                else:
                    means = means.reshape((-1,))
                    stds = stds.reshape((-1,))
                    if current_best is not None:
                        current_best = current_best.reshape((1,))
                output_to_mean_std[output_model] = {'mean': means, 'std': stds}
            # Compute the acquisition function value
            fvals = self._compute_heads(output_to_mean_std, current_best)
            # Average over fantasies if there are any
            if num_fantasies > 1:
                fvals = np.mean(fvals, axis=1)
            fvals_list.append(fvals)

        return np.mean(fvals_list, axis=0)

    def compute_acq_with_gradient(self, input: np.ndarray, model: Optional[OutputSurrogateModel] = None) -> \
            Tuple[float, np.ndarray]:
        if model is None:
            model = self.model
        if isinstance(model, SurrogateModel):
            self.model = dictionarize_objective(model)
        output_to_predictions = self._map_outputs_to_predictions(input.reshape(1, -1))
        self.num_mcmc_samples = len(output_to_predictions[self.active_metric])  # a single sample means no MCMC
        current_best_list = self._get_current_best_for_active_metric()

        fvals = []  # this will contain the value of the acquisition function for each MCMC sample (if any)
        output_to_head_gradients = {output_name: [] for output_name in self.model_output_names}  # this dictionary will
        # map each output model to its dictionary of head gradients (one dictionary of head gradients per MCMC sample)
        # Loop over MCMC samples (if any)
        for predictions, current_best in zip(
                zip(*output_to_predictions.values()),
                current_best_list):
            output_model_to_mean_std = dict(zip(self.model_output_names, predictions))
            # Create a dictionary mapping each output model name to means, stds and current best,
            # all three in a shape that can be fed into self._compute_heads_and_gradient
            output_to_mean_std = {}
            for output_name, prediction in output_model_to_mean_std.items():
                mean = prediction['mean'].reshape((-1,))
                num_fantasies = mean.size
                std = prediction['std'].reshape((1,))
                if current_best is not None:
                    assert current_best.size == num_fantasies
                    current_best = current_best.reshape((-1,))
                output_to_mean_std[output_name] = {'mean': mean, 'std': std}
            head_result = self._compute_head_and_gradient(output_to_mean_std, current_best)
            fvals.append(np.mean(head_result.hvals))

            for output_name in self.model_output_names:
                # Fill up output_to_head_gradients, which maps each output_name to its head gradients. Head gradients
                # are a list: each list element is an MCMC sample and contains the head gradient of the mean and std.

                # Each head_gradient is composed for backward_gradient call. This involves reshaping. Also, we are
                # averaging over fantasies (if num_fantasies > 1), which is not done in _compute_head_and_gradient
                # (dh_dmean, dh_dstd have the same shape as mean), so we have to divide by num_fantasies.
                output_mean = output_model_to_mean_std[output_name]['mean']
                output_std = output_model_to_mean_std[output_name]['std']
                num_fantasies = output_mean.size
                head_gradient = {'mean': head_result.dh_dmean[output_name].reshape(
                    output_mean.shape) / num_fantasies,
                                 'std': np.array([np.mean(head_result.dh_dstd[output_name])]).reshape(
                                     output_std.shape)}
                output_to_head_gradients[output_name].append(head_gradient)

        fval = np.mean(fvals)  # Averaging over MCMC samples (if any)
        gradient = 0.0
        # Sum up the gradients coming from each output model
        for output_name, output_model in model.items():
            # Gradients are computed by the model
            gradient_list = output_model.backward_gradient(input, output_to_head_gradients[output_name])
            output_gradient = np.mean(gradient_list, axis=0)  # Averaging over MCMC samples (if any)
            gradient += output_gradient
        return fval, gradient

    def _map_outputs_to_predictions(self, inputs):
        """
        Returns a dictionary mapping each output to the predictive mean and std at the given inputs.
        """
        output_to_predictions = {}
        for output_name, output_model in self.model.items():
            assert 'mean' in output_model.keys_predict()
            assert 'std' in output_model.keys_predict()
            output_to_predictions[output_name] = output_model.predict(inputs)
        return output_to_predictions

    def _get_current_best_for_active_metric(self):
        """
        Returns a list of current best, one per MCMC sample (if head requires the current best).
        """
        if self._head_needs_current_best():
            current_best_list = self.model[self.active_metric].current_best()  # list of  1 x nf arrays
        else:
            current_best_list = [None] * self.num_mcmc_samples
        return current_best_list

    def _extract_active_metric_stats(self, output_to_mean_std):
        mean = output_to_mean_std[self.active_metric]['mean']
        stds = output_to_mean_std[self.active_metric]['std']
        return mean, stds

    @abstractmethod
    def _head_needs_current_best(self) -> bool:
        """
        :return: Is the current_best argument in _compute_head needed?
        """
        pass

    @abstractmethod
    def _compute_heads(
            self, output_to_mean_std: Dict[str, Dict[str, np.ndarray]],
            current_best: Optional[np.ndarray]) -> np.ndarray:
        """
        If mean has >1 columns, both std and current_best are supposed to be
        broadcasted. The return value has the same shape as mean.

        :param: output_to_mean_std_best: Dictionary mapping each output to a dictionary containing:
                                        mean: Predictive means, shape (n, nf) or (n,)
                                        std: Predictive stddevs, shape (n, 1) or (n,)
                current_best: Incumbent, shape (1, nf) or (1,)
        :return: h(means, stds, current_best), same shape as means
        """
        pass

    @abstractmethod
    def _compute_head_and_gradient(
            self, output_to_mean_std: Dict[str, Dict[str, np.ndarray]],
            current_best: Optional[np.ndarray]) -> HeadWithGradient:
        """
        Computes both head value and head gradients, for a single input.

        :param: output_to_mean_std: Dictionary mapping each output to a dictionary containing:
                                    mean: Predictive mean, shape (nf,)
                                    std: Predictive stddev, shape (1,)
                current_best: Incumbent, shape (nf,)
        :return: HeadWithGradient containing hvals and, for each output model, dh_dmean, dh_dstd.
                 All HeadWithGradient values have the same shape as mean

        """
        pass


class AcquisitionWithMultiModelCurrentBest(MeanStdAcquisitionFunction, ABC):
    """
    Abstract acquisition function class for multi-output acquisition functions
    requiring predictions from all output models to compute the current best.

    For instance, CEIAcquisitionFunction uses the output from the constraint model
    to filter out the infeasible candidates before computing the current best.
    """
    def _map_models_to_candidate_predictive_means(self):
        # Create a dictionary mapping each output model name to the predictive means at the current candidates
        models_to_means = {model_name: model.predict_mean_current_candidates()
                           for model_name, model
                           in self.model.items()}
        expected_length = len(next(iter(models_to_means.values())))
        assert all(len(predicted_means) == expected_length
                   for predicted_means
                   in models_to_means.values()), \
            'All models must have the same number of MCMC samples.'
        return models_to_means

    @abstractmethod
    def _get_current_best_for_active_metric(self):
        """
        Returns a list of current best, one per MCMC sample (if head requires the current best).

        Overrides the parent method to use the predictions from all models to identify the current best.
        """
        pass



import multiprocessing as mp

from autogluon.core.scheduler.remote import RemoteManager
from autogluon.core.scheduler.resource import DistributedResourceManager


class TaskManagers(object):

    def __init__(self):
        self._resource_manager = None
        self._remote_manager = None
        self.lock = mp.Lock()

    @property
    def resource_manager(self):
        if self._resource_manager is None:
            self._resource_manager = DistributedResourceManager()
        return self._resource_manager

    @property
    def remote_manager(self):
        if self._remote_manager is None:
            self._remote_manager = RemoteManager()
        return self._remote_manager

    # --------- Delegates ------------

    def register_dist_ip_addrs(self, dist_ip_addrs):
        if dist_ip_addrs is None:
            dist_ip_addrs = []
        self.remote_manager.add_remote_nodes(dist_ip_addrs)
        self.resource_manager.add_remote(self.remote_manager.get_remotes())

    def add_remote(self, ip_addrs):
        """Add remote nodes to the scheduler computation resource.
        """
        ip_addrs = [ip_addrs] if isinstance(ip_addrs, str) else ip_addrs
        with self.lock:
            remotes = self.remote_manager.add_remote_nodes(ip_addrs)
            self.resource_manager.add_remote(remotes)

    def upload_files(self, files, **kwargs):
        """Upload files to remote machines, so that they are accessible by import or load.
        """
        self.remote_manager.upload_files(files, **kwargs)

    def request_resources(self, resources):
        self.resource_manager._request(resources)

    def release_resources(self, resources):
        self.resource_manager._release(resources)

    def check_availability(self, resources):
        return self.resource_manager.check_availability(resources)



import copy
import logging
import os
import time
from abc import abstractmethod

from numpy import ndarray
from pandas import DataFrame, Series

from ...utils.exceptions import TimeLimitExceeded

logger = logging.getLogger(__name__)


class AbstractFoldFittingStrategy(object):

    @abstractmethod
    def schedule_fold_model_fit(self, model_base, fold_ctx, kwargs):
        """
        Schedule fold model training. By design this part is supposed to be 'lazy' evaluator no actual training is performed here.
        Distributed fitters will handle jobs scheduling here.
        """
        pass

    @abstractmethod
    def after_all_folds_scheduled(self):
        """
        Method is called when all the folds are scheduled.
        Local fitters will perform training here.
        Distributed fitters will handle job handles and results retrieval here.
        """
        pass


class SequentialLocalFoldFittingStrategy(AbstractFoldFittingStrategy):
    """
    This strategy fits the folds locally in a sequence.
    """

    def __init__(self, bagged_ensemble_model, X: DataFrame, y: Series, sample_weight, time_limit: float, time_start: float,
                 models: list, oof_pred_proba: ndarray, oof_pred_model_repeats: ndarray, save_folds: bool):
        self.X = X
        self.y = y
        self.sample_weight = sample_weight
        self.time_limit = time_limit
        self.time_start = time_start
        self.models = models
        self.oof_pred_proba = oof_pred_proba
        self.oof_pred_model_repeats = oof_pred_model_repeats
        self.bagged_ensemble_model = bagged_ensemble_model
        self.jobs = []
        self.save_folds = save_folds

    def schedule_fold_model_fit(self, model_base, fold_ctx, kwargs):
        self.jobs.append([model_base, fold_ctx, kwargs])

    def after_all_folds_scheduled(self):
        for job in self.jobs:
            self._fit_fold_model(*job)

    def _fit_fold_model(self, model_base, fold_ctx, kwargs):
        time_start_fold = time.time()
        time_limit_fold = self._get_fold_time_limit(fold_ctx)
        fold_model = self._fit(model_base, time_start_fold, time_limit_fold, fold_ctx, kwargs)
        fold_model, pred_proba = self._predict_oof(fold_model, fold_ctx)
        self._update_bagged_ensemble(fold_model, pred_proba, fold_ctx)

    def _get_fold_time_limit(self, fold_ctx):
        _, folds_finished, folds_left, folds_to_fit, _, _ = self._get_fold_properties(fold_ctx)
        time_elapsed = time.time() - self.time_start
        if self.time_limit is not None:
            time_left = self.time_limit - time_elapsed
            required_time_per_fold = time_left / folds_left
            time_limit_fold = required_time_per_fold * 0.8
            if folds_finished > 0:
                expected_time_required = time_elapsed * folds_to_fit / folds_finished
                expected_remaining_time_required = expected_time_required * folds_left / folds_to_fit
                if expected_remaining_time_required > time_left:
                    raise TimeLimitExceeded
            if time_left <= 0:
                raise TimeLimitExceeded
        else:
            time_limit_fold = None
        return time_limit_fold

    def _update_bagged_ensemble(self, fold_model, pred_proba, fold_ctx):
        _, val_index = fold_ctx['fold']
        model_to_append = fold_model
        if not self.save_folds:
            fold_model.model = None
        if self.bagged_ensemble_model.low_memory:
            self.bagged_ensemble_model.save_child(fold_model, verbose=False)
            model_to_append = fold_model.name
        self.models.append(model_to_append)
        self.oof_pred_proba[val_index] += pred_proba
        self.oof_pred_model_repeats[val_index] += 1
        self.bagged_ensemble_model._add_child_times_to_bag(model=fold_model)

    def _fit(self, model_base, time_start_fold, time_limit_fold, fold_ctx, kwargs):
        fold, folds_finished, folds_left, folds_to_fit, is_last_fold, model_name_suffix = self._get_fold_properties(fold_ctx)
        train_index, val_index = fold
        X_fold, X_val_fold = self.X.iloc[train_index, :], self.X.iloc[val_index, :]
        y_fold, y_val_fold = self.y.iloc[train_index], self.y.iloc[val_index]
        fold_model = copy.deepcopy(model_base)
        fold_model.name = f'{fold_model.name}{model_name_suffix}'
        fold_model.set_contexts(self.bagged_ensemble_model.path + fold_model.name + os.path.sep)
        kwargs_fold = kwargs.copy()
        if self.sample_weight is not None:
            kwargs_fold['sample_weight'] = self.sample_weight[train_index]
            kwargs_fold['sample_weight_val'] = self.sample_weight[val_index]
        fold_model.fit(X=X_fold, y=y_fold, X_val=X_val_fold, y_val=y_val_fold, time_limit=time_limit_fold, **kwargs_fold)
        fold_model.fit_time = time.time() - time_start_fold
        return fold_model

    def _predict_oof(self, fold_model, fold_ctx):
        time_train_end_fold = time.time()
        fold, folds_finished, folds_left, folds_to_fit, is_last_fold, model_name_suffix = self._get_fold_properties(fold_ctx)
        _, val_index = fold
        X_val_fold = self.X.iloc[val_index, :]
        y_val_fold = self.y.iloc[val_index]
        if self.time_limit is not None:  # Check to avoid unnecessarily predicting and saving a model when an Exception is going to be raised later
            if not is_last_fold:
                time_elapsed = time.time() - self.time_start
                time_left = self.time_limit - time_elapsed
                expected_time_required = time_elapsed * folds_to_fit / (folds_finished + 1)
                expected_remaining_time_required = expected_time_required * (folds_left - 1) / folds_to_fit
                if expected_remaining_time_required > time_left:
                    raise TimeLimitExceeded
        pred_proba = fold_model.predict_proba(X_val_fold)
        fold_model.predict_time = time.time() - time_train_end_fold
        fold_model.val_score = fold_model.score_with_y_pred_proba(y=y_val_fold, y_pred_proba=pred_proba)
        fold_model.reduce_memory_size(remove_fit=True, remove_info=False, requires_save=True)
        if not self.bagged_ensemble_model.params.get('save_bag_folds', True):
            fold_model.model = None
        return fold_model, pred_proba

    @staticmethod
    def _get_fold_properties(fold_ctx):
        fold, folds_finished, folds_left, folds_to_fit, is_last_fold, model_name_suffix = [
            fold_ctx[f] for f in ['fold', 'folds_finished', 'folds_left', 'folds_to_fit', 'is_last_fold', 'model_name_suffix']
        ]
        return fold, folds_finished, folds_left, folds_to_fit, is_last_fold, model_name_suffix



import autograd.numpy as anp
import autograd.scipy.linalg as aspl
from autograd.extend import primitive, defvjp
import numpy as np
import scipy.linalg as spl
import logging
import math

logger = logging.getLogger(__name__)

__all__ = ['AddJitterOp',
           'flatten_and_concat',
           'cholesky_factorization']


INITIAL_JITTER_FACTOR = 1e-9
JITTER_GROWTH = 10.
JITTER_UPPERBOUND_FACTOR = 1e3


def flatten_and_concat(x: anp.ndarray, sigsq_init: anp.ndarray):
    return anp.append(anp.reshape(x, (-1,)), sigsq_init)


@primitive
def AddJitterOp(inputs: np.ndarray, initial_jitter_factor=INITIAL_JITTER_FACTOR,
                jitter_growth=JITTER_GROWTH, debug_log='false'):      
    """
    Finds smaller jitter to add to diagonal of square matrix to render the
    matrix positive definite (in that linalg.potrf works).

    Given input x (positive semi-definite matrix) and sigsq_init (nonneg
    scalar), find sigsq_final (nonneg scalar), so that:
        sigsq_final = sigsq_init + jitter, jitter >= 0,
        x + sigsq_final * Id positive definite (so that potrf call works)
    We return the matrix x + sigsq_final * Id, for which potrf has not failed.

    For the gradient, the dependence of jitter on the inputs is ignored.

    The values tried for sigsq_final are:
        sigsq_init, sigsq_init + initial_jitter * (jitter_growth ** k),
        k = 0, 1, 2, ...,
        initial_jitter = initial_jitter_factor * max(mean(diag(x)), 1)

    Note: The scaling of initial_jitter with mean(diag(x)) is taken from GPy.
    The rationale is that the largest eigenvalue of x is >= mean(diag(x)), and
    likely of this magnitude.

    There is no guarantee that the Cholesky factor returned is well-conditioned
    enough for subsequent computations to be reliable. A better solution
    would be to estimate the condition number of the Cholesky factor, and to add
    jitter until this is bounded below a threshold we tolerate. See

        Higham, N.
        A Survey of Condition Number Estimation for Triangular Matrices
        MIMS EPrint: 2007.10

    Algorithm 4.1 could work for us.
    """
    assert initial_jitter_factor > 0. and jitter_growth > 1.
    n_square = inputs.shape[0] - 1
    n = int(math.sqrt(n_square))
    assert n_square % n == 0 and n_square // n == n, "x must be square matrix, shape (n, n)"
    x, sigsq_init = np.reshape(inputs[:-1], (n, -1)), inputs[-1]
    
    def _get_constant_identity(x, constant):
        n, _ = x.shape
        return np.diag(np.ones((n,)) * constant)

    def _get_jitter_upperbound(x):
        # To define a safeguard in the while-loop of the forward,
        # we define an upperbound on the jitter we can reasonably add
        # the bound is quite generous, and is dependent on the scale of the input x
        # (the scale is captured via the trace of x)
        # the primary goal is avoid any infinite while-loop.
        return JITTER_UPPERBOUND_FACTOR * max(1., np.mean(np.diag(x)))

    jitter = 0.
    jitter_upperbound = _get_jitter_upperbound(x)
    must_increase_jitter = True
    x_plus_constant = None
    
    while must_increase_jitter and jitter <= jitter_upperbound:
        try:
            x_plus_constant = x + _get_constant_identity(
                x, sigsq_init + jitter)
            # Note: Do not use np.linalg.cholesky here, this can cause
            # locking issues
            L = spl.cholesky(x_plus_constant, lower=True)
            must_increase_jitter = False
        except spl.LinAlgError:
            if debug_log == 'true':
                logger.info("sigsq = {} does not work".format(
                    sigsq_init + jitter))
            if jitter == 0.0:
                jitter = initial_jitter_factor * max(1., np.mean(np.diag(x)))
            else:
                jitter = jitter * jitter_growth

    assert not must_increase_jitter, "The jitter ({}) has reached its upperbound ({}) while the Cholesky of the input matrix still cannot be computed.".format(jitter, jitter_upperbound)
    
    if debug_log == 'true':
        logger.info("sigsq_final = {}".format(sigsq_init + jitter))

    return x_plus_constant


def AddJitterOp_vjp(
        ans: np.ndarray, inputs: np.ndarray,
        initial_jitter_factor=INITIAL_JITTER_FACTOR, jitter_growth=JITTER_GROWTH,
        debug_log='false'):
    return lambda g: anp.append(anp.reshape(g, (-1,)), anp.sum(anp.diag(g)))


defvjp(AddJitterOp, AddJitterOp_vjp)


@primitive
def cholesky_factorization(a):
    """
    Replacement for autograd.numpy.linalg.cholesky. Our backward (vjp) is
    faster and simpler, while somewhat less general (only works if
    a.ndim == 2).

    See https://arxiv.org/abs/1710.08717 for derivation of backward (vjp)
    expression.
    
    :param a: Symmmetric positive definite matrix A
    :return: Lower-triangular Cholesky factor L of A
    """
    # Note: Do not use np.linalg.cholesky here, this can cause locking issues
    return spl.cholesky(a, lower=True)


def copyltu(x):
    return anp.tril(x) + anp.transpose(anp.tril(x, -1))


def cholesky_factorization_backward(l, lbar):
    abar = copyltu(anp.matmul(anp.transpose(l), lbar))
    abar = anp.transpose(aspl.solve_triangular(l, abar, lower=True, trans='T'))
    abar = aspl.solve_triangular(l, abar, lower=True, trans='T')
    return 0.5 * abar


def cholesky_factorization_vjp(l, a):
    return lambda lbar: cholesky_factorization_backward(l, lbar)


defvjp(cholesky_factorization, cholesky_factorization_vjp)
